-----------------------------------------------------------------------------
-- |
-- Module      :  Conjugacy
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Implementation of conjugacy simplification
--
-----------------------------------------------------------------------------

module Conjugacy where

import Core

import Data.List
import Data.Maybe
import Control.Monad
import Debug.Trace
import qualified UnLift

-- to allow for user-defined conjugacy rules
data Conjugate = Conjugate { prior :: Id
                           , likelihood :: Id
                           , posterior  :: Id   -- allow distribution switching, eg., DirSym -> Dir
                           , matching   :: Int  -- which expression index in the posterior has to match the prior parameter
                           , idxdelta   :: Int  -- difference in # of indices between prior on posterior.  eg. Gam(a_k) vs Dir(a) --> idxdelta = 1 because
                                                --              there should be one more index on the prior
                           , suffstats  :: [SExpr] -> [Expr] -> Lik -> [Expr]
                           }

-- type of likelihoods
type Lik = [([Range],Dist,[Expr])]
          -- means prod_(rl,d,el) prod_rl d^(prod el)

-- standard rules
conjugacyRules =
    [ Conjugate { prior      = "Dir"
                , likelihood = "Mult"
                , posterior  = "Dir"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjDirMult
                }
    , Conjugate { prior      = "DirSym"
                , likelihood = "Mult"
                , posterior  = "Dir"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjDirSymMult
                }
    , Conjugate { prior      = "NorMV"
                , likelihood = "NorMV"
                , posterior  = "NorMV"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const (conjNorNor True)
                }
    , Conjugate { prior      = "Nor"
                , likelihood = "Nor"
                , posterior  = "Nor"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const (conjNorNor False)
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "Exp"
                , posterior  = "Gam"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjGamExp
                }
    , Conjugate { prior      = "Bet"
                , likelihood = "Bin"
                , posterior  = "Bet"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjBetBin
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "DirSym"
                , posterior  = "Gam"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjGamDirSym
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "Dir"
                , posterior  = "Gam"
                , matching   = 0
                , idxdelta   = 1
                , suffstats  = conjGamDir
                }
    , Conjugate { prior      = "Gam"
                , likelihood = "Poi"
                , posterior  = "Gam"
                , matching   = 0
                , idxdelta   = 0
                , suffstats  = const conjGamPoi
                }
    , Conjugate { prior      = "IG"
                , likelihood = "NorMV"
                , posterior  = "IG"
                , matching   = 1
                , idxdelta   = 0
                , suffstats  = const (conjGamNor True)
                }
    , Conjugate { prior      = "IG"
                , likelihood = "Nor"
                , posterior  = "IG"
                , matching   = 1
                , idxdelta   = 0
                , suffstats  = const (conjGamNor False)
                }
    ]


conjDirMult [al,len] lik 
--  | trace (show ("conjDirMult:", al,len,lik)) True 
  = [foldr conj al lik, len]
  where
    conj (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd' v0 el) rl]
      where v0 = Func "IDR" [Var v, len]
    
conjDirSymMult [al,len] lik = [foldr conj al0 lik, len]
  where
    al0 = Func "vec" [al, Con (I 1), len]
    conj (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd' v0 el) rl]
      where v0 = Func "IDR" [Var v, len]

conjGamDirSym [a,b] lik = [a, b']
  where
    b' = Func "/" [one, Func "-" [Func "/" [one, b], foldr conj (Con (R 0)) lik]]
    one   = Con (R 1)
    conj (rl, Dist v _ [_,dim], el) d0 = 
      Func "+" [d0, foldr makeLogSum (foldr makeProd (sumlog dim v) el) rl]
    makeLogSum r e = SP False r e ""
    sumlog dim v0@(V v vl) = SP False (Range "cgds" (Con (I 1)) dim) (Func "log" [Func ".*" [norm,v1]]) ""
      where 
        norm = Func "/" [one, SP False (Range "cgds_norm" (Con (I 1)) dim) v1' ""]
        v1   = Var (V v (vl ++ [Id "cgds"      Nothing]))
        v1'  = Var (V v (vl ++ [Id "cgds_norm" Nothing]))

conjGamDir idx0 [a,b] lik = [a, b']
  where
    b' = Func "/" [one, Func "-" [Func "/" [one, b], foldr conj (Con (R 0)) lik]]
    one   = Con (R 1)
    conj (rl, Dist v _ [_,dim], el) d0 = 
      Func "+" [d0, foldr makeLogSum (foldr makeProd (sumlog dim v) el) rl]
    makeLogSum r e = SP False r e ""
    sumlog dim v0@(V v vl) = Func "log" [Func ".*" [norm,v1]]
      where 
        norm = Func "/" [one, SP False (Range "cgds_norm" (Con (I 1)) dim) v1' ""]
        v1   = Var (V v (vl ++ idx0)) -- [Id "cgds"      Nothing]))
        v1'  = Var (V v (vl ++ [Id "cgds_norm" Nothing]))

makeSum  r e  = SP False r e ""
makeProd  e0 e = Func  "*" [e0, e]
makeProd' e0 e = Func ".*" [e0, e]

conjGamNor isMultivar [al,be] lik = [al', be']
  where
    al' = Func "+" [al, Func "/" [n   , Con (R 2)]]
    be' = Func "+" [be, Func "/" [sumx, Con (R 2)]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ (mu:_), el) = foldr makeSum (foldr makeProd' (d mu $ Var v) el) rl
    makePost False (rl, Dist v _ _     , el) = foldr makeSum (foldr makeProd  one            el) rl

    d mu e 
      | isMultivar = Func "sqrDiff" [e,mu]
      | otherwise  = Func "*" [Func "-" [e,mu], Func "-" [e,mu]]
--      | isMultivar = Func "sum" [Func "*" [Func "-" [e,mu], Func "-" [e,mu]]]
--      | otherwise  = Func "*" [Func "-" [e,mu], Func "-" [e,mu]]


conjNorNor isMultivar (mu0:si0:rest) lik 
  | length all_si /= 1 = error ("Conjugacy.conjNorNor: not all likelihood variances are the same... this is fixable, but not yet implemented")
  | length all_mu /= 1 = error ("Conjugacy.conjNorNor: not all likelihood means are the same... this is fixable, but not yet implemented")
  | otherwise = (mu' : rho' : rest)
  where
    all_si = nub [si | (_,Dist _ _ (_:si:_),_) <- lik]
    all_mu = nub [si | (_,Dist _ _ (mu:_:_),_) <- lik]
    si     = head all_si

    --   = (1/si0 + n/si)^{-1}; n = sum_...
    rho' = Func "^" [Func "+" [Func "/" [one, si0]
                              ,Func "*" [Func "/" [one, si], n]]
                    ,Con (R (-1))]

    times = if isMultivar then ".*" else "*"

    --   = rho' * (mu0/si0 + (1/si) sum x)
    mu'  = Func times [rho'
                      ,Func "+" [Func times [Func "/" [one, si0], mu0]
                                ,Func times [Func "/" [one, si ], sumx]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    -- helpers
    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl

conjGamExp [al,be] lik = [al',be']
  where
    al' = Func "+" [al, n]
    be' = Func "/" [be, Func "+" [one, Func "*" [be, sumx]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl

conjGamPoi [al,be] lik = [al', be']
  where
    al' = Func "+" [al, sumx]
    be' = Func "/" [be, Func "+" [one, Func "*" [be, n]]]

    sumx = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost True ) lik
    n    = foldr1 (\a b -> Func "+" [a,b]) $ map (makePost False) lik

    one = Con (R 1)

    makePost True  (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd' (Var v) el) rl
    makePost False (rl, Dist v _ _, el) = foldr makeSum (foldr makeProd  one     el) rl


conjBetBin [al,be] lik = [foldr (conj True ) al lik
                         ,foldr (conj False) al lik]
  where
    conj b (rl, Dist v _ _, el) d0 = Func "+" [d0, foldr makeSum (foldr makeProd (Con (I 1)) (if not b then map notit el else el)) rl]
    notit e = Func "not" [e]

simplify :: Core -> Core
simplify (Core l) = Core [(v,applyConjugacy conjugacyRules d,r) | (v,d,r) <- l]

applyConjugacy :: [Conjugate] -> Dist -> Dist
applyConjugacy rules d 
  | null pr_lik = d
  | otherwise   = applyConjugacy rules post
  where
    pr_lik = [ (priorDist0, prior, (idx0,lik,rest))
                   | (priorDist0,priorConj) <- findPrior conjugacyRules d
                   , let prior = (stripPrior $ UnLift.unlift priorDist0, priorConj)
--                   , trace (show ("prior=", fst prior)) True
                   , let (idx0,lik,rest) = findLikelihood prior d
                   , not (null lik)
                   , isJust rest
             ]
    (pr0,pr,lik) = head pr_lik
    post     = computePosterior pr0 pr lik

stripPrior (Prod _ d _) = stripPrior d
stripPrior (a :^: _) = stripPrior a
stripPrior d = d

findPrior :: [Conjugate] -> Dist ->  [(Dist, Conjugate)]
findPrior rules d@(Dist _ id _) = [(d,r) | r <- rules, prior r == id]
findPrior rules (a :*: b) = findPrior rules a `mplus`  findPrior rules b
findPrior rules (a :^: e) = [(d :^: e, c) | (d,c) <- findPrior rules a]
findPrior rules (Prod r d spid) = [(Prod r d spid, c) | (d,c) <- findPrior rules d]

findLikelihood :: (Dist, Conjugate) -> Dist -> ([SExpr], Lik, Maybe Dist) -- second is what remains
findLikelihood (Dist priorParam _ _, conj) d@(Dist _ id el)
  | id == likelihood conj &&
    matching conj < length el =
    case matchVariable (idxdelta conj) priorParam (el !! matching conj) of
      Nothing  -> ([], [], Just d)
      Just (idx0,el') -> (idx0, [([],d,el')], Nothing)
  | otherwise = ([], [], Just d)
findLikelihood dc (a :*: b) = (idx0A++idx0B,likA++likB, restA `distTimes` restB)
  where (idx0A,likA,restA) = findLikelihood dc a
        (idx0B,likB,restB) = findLikelihood dc b
findLikelihood dc (a :^: e) = (idx0, map (\ (r,d,el) -> (r,d,e:el)) lik, liftM (:^:e) rest)
  where (idx0,lik,rest) = findLikelihood dc a
findLikelihood dc (Prod range d spid) = (idx0, map (\ (r,d,el) -> (range:r,d,el)) lik, liftM (\d -> Prod range d spid) rest)
  where (idx0,lik,rest) = findLikelihood dc d

matchVariable 0 v e | Var v == e = Just ([], [])
matchVariable delta (V pid pse) (Var (V eid ese))
  | pid == eid && length pse == length ese + delta = Just (take delta pse, zipWith (\p e -> Func "=" [sexprToExpr p,sexprToExpr e]) (drop delta pse) ese)
matchVariable _ _ _ = Nothing

{-
matchVariable v e | Var v == e = Just []
matchVariable (V pid pse) (Var (V eid ese))
  | pid == eid && length pse == length ese = Just (zipWith (\p e -> Func "=" [sexprToExpr p,sexprToExpr e]) pse ese)
matchVariable _ _ = Nothing
-}

computePosterior :: Dist -> (Dist, Conjugate) -> ([SExpr], Lik, Maybe Dist) -> Dist
computePosterior _ _ (_,_,Nothing) = error ("Conjugacy.computePosterior: Maybe Dist is Nothing; at least the prior should be there")
computePosterior prior0 (prior@(Dist id0 _ el0), conj) (idx0, lik, Just rest) 
--  | trace (show ("computePosterior", prior0, prior, idx0, lik, Just rest)) True
  = fromJust ((rest `distWithout` prior0) `distTimes` Just (Dist id0 (posterior conj) ((suffstats conj) idx0 el0 lik)))
