-----------------------------------------------------------------------------
-- |
-- Module      :  Reindex
-- Project     :  HBC
-- Copyright   :  (c) Hal Daume III
-- License     :  Open
-- 
-- Maintainer  :  Hal Daume III <me@hal3.name>
-- Stability   :  stable
-- Portability :  portable
--
-- Implementation of the reindexing of RHSs to account for different
-- appearances of the variable we care about.  We assume that this step 
-- is performed *after* lifting; this means that (a) we have to handle 
-- lifts (bad), and (b) we can assume all INDs are semi-simple (good).
--
-----------------------------------------------------------------------------

module Reindex(reindex) where

import Decl
import Util
import Data.Maybe
import Debug.Trace
import Data.List
import Debug.Trace

reindex :: DECL -> DECL -> EXPR -> ((VAR, [EXPR]), RHS, [RANGE])
reindex (DECL (LHS v0 ix0) _ _) (DECL (LHS v1 ix1) (RHS dist el lift) ranges) e0 
--  | trace ("ixMap=" ++ show ixMap) False = undefined
  | otherwise = ((v1, map (applyMap ixMap.VAR) ix1), rhs',ranges')
  where
    ixMap = 
      case e0 of
        VAR v | null (ix0 `intersect` ix1) -> []   -- TODO: not sure if this is correct!
        IND v ix {- | length ix == length ix0 -} -> zipWith findIndexMap ix0 ix
        _ -> error ("Reindex.reindex: cannot compute index map (" ++ show v0 ++ ") for " ++ show ix0 ++ " and (" ++ show v1 ++ ") " ++ show e0 ++ " versus " ++ show (v1,ix1))

    rhs'    = RHS dist (map (applyMap ixMap) el) (map (applyMapLift ixMap) lift)
    ranges' = map (applyMapRange ixMap) ranges


-- simple examples: 
--    findIndexMap n m --> (n,m)
--    findIndexMap n n+1 --> (n, n-1)
--    findIndexMap n m+1 --> (m, n-1)
findIndexMap :: VAR -> EXPR -> (VAR, EXPR)
findIndexMap v0 e = findIndexMap' e (VAR v0)
  where
    findIndexMap' (VAR v) e0 = (v, e0)
    findIndexMap' (FN "exp" [a]) e0 = findIndexMap' a (FN "log" [e0])
    findIndexMap' (FN "log" [a]) e0 = findIndexMap' a (FN "exp" [e0])
    findIndexMap' f@(FN op  [a,b]) e0
--      | trace (show ("findIndexMap'", f, e0, hasVar a, hasVar b, op)) False = undefined
      | hasVar a && hasVar b        = error ("Reindex.findIndexMap':: both L and R have variables in " ++ show f)
      | not (hasVar a || hasVar b)  = error ("Reindex.findIndexMap':: neither L nor R have variables in " ++ show f)
      | hasVar a && op == "+"       = findIndexMap' a (FN "-" [e0,b])
      | hasVar b && op == "+"       = findIndexMap' b (FN "-" [e0,a])
      | hasVar a && op == "-"       = findIndexMap' a (FN "+" [e0,b])
      | hasVar b && op == "-"       = findIndexMap' b (FN "+" [a,e0])
      | otherwise                   = error ("Reindex.findIndexMap':: cannot invert function in " ++ show f)
      where
        hasVar = not . null . exprIds
    findIndexMap' v        e0      = error ("Reindex.findIndexMap':: cannot map " ++ show v)

applyMap :: [(VAR,EXPR)] -> EXPR -> EXPR
applyMap m (VAR v) = fromMaybe (VAR v) $ lookup v m
applyMap m (IND v el)
  | v `elem` map fst m = error ("Reindex.applyMap: map contains indexed variable '" ++ show v ++ "'; maybe lifting failed")
  | otherwise          = IND v $ map (applyMap m) el
applyMap m (FN f el) = FN f $ map (applyMap m) el
applyMap m (SUMPP b r e) = SUMPP b (applyMapRange m r) (applyMap m e)
applyMap m (CONST c) = CONST c

applyMapRange m (RANGE v lo hi) = RANGE (v {- fromMaybe v $ lookup v m -}) (applyMap m lo) (applyMap m hi)  -- todo: fix this; apply change in index to ranges!
--  | v `elem` map fst m = error ("Reindex.applyMapRange: map contains range variable " ++ show v)
--  | otherwise          = RANGE v (applyMap m lo) (applyMap m hi)

--applyMapLift m (VAR v,e) | v `elem` map fst m = (VAR "?", applyMap
applyMapLift m (v,e) = (applyMap m v, applyMap m e)
--  | v `elem` map fst m = error ("Reindex.applyMapLift: map contains lifted variable " ++ show v)
--  | otherwise          = (v, applyMap m e)
